import os
import json
import time
import requests
import numpy as np
import pandas as pd
import websocket
import threading
from datetime import datetime, timedelta
from sklearn.preprocessing import StandardScaler
from tensorflow.keras.models import load_model
import pickle
import logging
import sys
from Crypto.Cipher import AES
import base64
from settings import settings  # Import the settings from settings.py

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Initialize paths
data_dir = settings.get('data_path', os.path.join(os.path.dirname(__file__), '..', 'data'))
stock_model_path = os.path.join(data_dir, 'models', 'lstm_model_stock.h5')
stock_scaler_path = os.path.join(data_dir, 'models', 'scaler_stock.pkl')
all_model_path = os.path.join(data_dir, 'models', 'lstm_model_all.h5')
all_scaler_path = os.path.join(data_dir, 'models', 'scaler_all.pkl')

# Check if the necessary settings are loaded
required_keys = ['appkey', 'appsecret', 'token_path', 'trading_url']
missing_keys = [key for key in required_keys if not settings.get(key)]

if missing_keys:
    raise ValueError(f"Missing required settings: {', '.join(missing_keys)}")

# Add config directory to sys.path
current_dir = os.path.dirname(os.path.abspath(__file__))
config_dir = os.path.abspath(os.path.join(current_dir, '..', '..', 'config'))
if config_dir not in sys.path:
    sys.path.insert(0, config_dir)

# 1. 모델 및 스케일러 로드
def load_lstm_model_and_scaler(model_path, scaler_path):
    try:
        model = load_model(model_path)
        with open(scaler_path, 'rb') as f:
            scaler = pickle.load(f)
        return model, scaler
    except Exception as e:
        logging.error(f"Failed to load model or scaler: {e}")
        raise

# 2. 실시간 데이터 수집
def get_stock_codes_from_csv(file_path):
    """CSV 파일에서 종목 코드를 읽어옵니다."""
    try:
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"CSV 파일을 찾을 수 없습니다: {file_path}")
        
        df = pd.read_csv(file_path, dtype={'code': str}, usecols=['code'])
        
        if df.empty:
            raise ValueError("CSV 파일이 비어 있습니다.")
        
        return df['code'].tolist()
        
    except Exception as e:
        logging.error(f"CSV 파일에서 종목 코드를 읽는 중 오류가 발생했습니다: {e}")
        return []

def get_valid_token():
    """유효한 토큰을 가져옵니다. 없거나 만료된 경우 새로 갱신합니다."""
    token_path = settings['token_path']
    try:
        with open(token_path, 'r') as file:
            token_data = json.load(file)
            if is_token_valid(token_data):
                return token_data['access_token']
            else:
                return refresh_token()
    except (FileNotFoundError, json.JSONDecodeError):
        return refresh_token()

def is_token_valid(token_data):
    """토큰이 여전히 유효한지 확인합니다."""
    expiration = datetime.strptime(token_data['expiration'], '%Y-%m-%d %H:%M:%S')
    return expiration > datetime.now()

def refresh_token():
    """만료되었거나 유효하지 않은 경우 토큰을 갱신합니다."""
    url = f"{settings['trading_url']}/oauth2/tokenP"
    data = {
        'grant_type': 'client_credentials',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret']
    }
    headers = {'Content-Type': 'application/json'}
    try:
        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            response_data = response.json()
            token_data = {
                'access_token': response_data['access_token'],
                'expiration': (datetime.now() + timedelta(seconds=response_data['expires_in'])).strftime('%Y-%m-%d %H:%M:%S')
            }
            with open(settings['token_path'], 'w') as file:
                json.dump(token_data, file)
            return token_data['access_token']
        else:
            raise Exception(f"토큰 갱신 실패: {response.text}")
    except requests.exceptions.RequestException as e:
        raise Exception(f"토큰 갱신 실패: {e}")

def decrypt_data(encrypted_data, key, iv):
    """AES256을 사용하여 데이터를 복호화합니다."""
    try:
        cipher = AES.new(key.encode('utf-8'), AES.MODE_CBC, iv.encode('utf-8'))
        decrypted = cipher.decrypt(base64.b64decode(encrypted_data))
        return decrypted.decode('utf-8').rstrip("\0")
    except Exception as e:
        logging.error(f"Failed to decrypt data: {e}")
        return None

def on_message(ws, message):
    """실시간 WebSocket 메시지를 처리합니다."""
    try:
        data = json.loads(message)
        if data.get('body', {}).get('msg1') == "SUBSCRIBE SUCCESS":
            logging.info("성공적으로 구독되었습니다.")
        elif data.get('body', {}).get('output', {}).get('iv'):
            iv = data['body']['output']['iv']
            key = data['body']['output']['key']
            encrypted_data = data['body']['output']['data']
            decrypted_data = decrypt_data(encrypted_data, key, iv)
            if decrypted_data:
                process_real_time_data(json.loads(decrypted_data))
        else:
            process_real_time_data(data)
    except json.JSONDecodeError:
        logging.error("WebSocket 메시지 디코딩 중 오류가 발생했습니다.")
    except KeyError as e:
        logging.error(f"필수 데이터가 누락되었습니다: {e}")

def on_error(ws, error):
    """WebSocket 오류를 처리합니다."""
    logging.error(f"WebSocket 오류 발생: {error}")

def on_close(ws, close_status_code, close_msg):
    """WebSocket 연결 종료를 처리합니다."""
    logging.warning(f"WebSocket이 종료되었습니다: {close_status_code} - {close_msg}")
    for attempt in range(3):
        logging.info(f"Reconnecting... Attempt {attempt + 1}")
        try:
            time.sleep(5 * (2 ** attempt))  # Exponential backoff
            run_websocket()
            break
        except Exception as e:
            logging.error(f"Reconnection attempt {attempt + 1} failed: {e}")
            if attempt == 2:
                logging.error("Max reconnection attempts reached. Exiting.")
                break

def on_open(ws):
    """WebSocket 연결이 열릴 때 구독 메시지를 전송합니다."""
    def run():
        approval_key = get_valid_token()
        stock_codes = get_stock_codes_from_csv(os.path.join(data_dir, 'stock_codes.csv'))

        for stock_code in stock_codes:
            subscribe_message = {
                "header": {
                    "approval_key": approval_key,
                    "custtype": "P",
                    "tr_type": "1",
                    "content-type": "utf-8"
                },
                "body": {
                    "input": {
                        "tr_id": "H0STASP0",
                        "tr_key": stock_code
                    }
                }
            }
            ws.send(json.dumps(subscribe_message))
            time.sleep(1)  # 각 요청 간에 딜레이 추가
    thread = threading.Thread(target=run)
    thread.start()

# 3. 데이터 수집 대기 및 예측 수행
def create_dataset(data, time_step):
    """LSTM 모델을 위한 데이터셋을 생성합니다."""
    X, y = [], []
    for i in range(len(data) - time_step):
        X.append(data[i:(i + time_step), 0])
        y.append(data[i + time_step, 0])
    return np.array(X), np.array(y)

def update_real_time_model(model, scaler, ticker, time_step=60):
    """새로운 데이터로 실시간 모델을 업데이트합니다."""
    real_time_data = get_real_time_data_with_retry(ticker)
    if real_time_data is None:
        return model
    
    if 'close' not in real_time_data.columns:
        real_time_data = real_time_data[['stck_prpr']].rename(columns={'stck_prpr': 'close'})
    
    if len(real_time_data) < time_step:
        logging.warning(f"주어진 time_step={time_step}에 대한 데이터가 충분하지 않습니다. 데이터 길이={len(real_time_data)}")
        return model
    
    real_time_data, _ = normalize_data(real_time_data)
    scaled_real_time_data = real_time_data['close'].values.reshape(-1, 1)
    
    X_real_time, y_real_time = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        logging.warning("데이터셋 생성 후 충분한 데이터 포인트가 없습니다.")
        return model
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], 1)
    
    model = update_lstm_model(pd.DataFrame(y_real_time), model, time_step)
    return model

def process_real_time_data(data):
    """실시간 데이터를 처리하여 예측을 수행합니다."""
    try:
        ticker = data['MKSC_SHRN_ISCD']
        ask_price_1 = data['ASKP1']
        bid_price_1 = data['BIDP1']
        accumulated_volume = data.get('ACML_VOL', 0)
        
        # 데이터 유효성 검사
        if ticker and ask_price_1 and bid_price_1 and accumulated_volume:
            # 최소 데이터가 모일 때까지 대기
            if accumulated_volume < 30:
                logging.info(f"{ticker}에 대한 데이터가 충분하지 않습니다. 더 많은 데이터를 기다립니다...")
                return

            real_time_predictions = real_time_prediction(stock_model, stock_scaler, ticker, time_step=30)
            if real_time_predictions is not None:
                logging.info(f"{ticker}에 대한 실시간 예측: {real_time_predictions}")
            else:
                logging.error(f"{ticker}에 대한 예측 실패.")
        else:
            logging.warning(f"{ticker}에 필요한 데이터가 누락되었습니다.")
    except KeyError as e:
        logging.error(f"WebSocket 데이터에서 누락된 키: {e}")

def real_time_prediction(model, scaler, ticker, time_step=60):
    """훈련된 모델을 사용하여 실시간 예측을 수행합니다."""
    real_time_data = get_real_time_data_with_retry(ticker)
    
    if real_time_data is None or len(real_time_data) < time_step:
        logging.warning(f"{ticker}에 대한 데이터가 충분하지 않습니다. 필요한 데이터: {time_step}, 받은 데이터: {len(real_time_data) if real_time_data is not None else 0}.")
        return None
    
    real_time_data = real_time_data[['stck_prpr']].rename(columns={'stck_prpr': 'close'})
    real_time_data, _ = normalize_data(real_time_data)
    scaled_real_time_data = real_time_data['close'].values.reshape(-1, 1)
    
    X_real_time, _ = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        logging.warning(f"{ticker}에 대한 데이터셋 생성 후 충분한 데이터 포인트가 없습니다.")
        return None
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], 1)
    
    predictions = model.predict(X_real_time)
    predictions = scaler.inverse_transform(predictions)
    return predictions

def real_time_prediction_all(model, scaler, ticker, time_step=60):
    """전체 데이터를 사용하는 모델을 통해 실시간 예측을 수행합니다."""
    real_time_data = get_real_time_data_with_retry(ticker)
    
    if real_time_data is None or len(real_time_data) < time_step:
        logging.warning(f"{ticker}에 대한 데이터가 충분하지 않습니다. 필요한 데이터: {time_step}, 받은 데이터: {len(real_time_data) if real_time_data is not None else 0}.")
        return None
    
    # 필요한 모든 특성을 포함하도록 실시간 데이터를 전처리합니다.
    real_time_data = preprocess_real_time_data_all(real_time_data)
    scaled_real_time_data = scaler.transform(real_time_data)
    
    X_real_time, _ = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        logging.warning(f"{ticker}에 대한 데이터셋 생성 후 충분한 데이터 포인트가 없습니다.")
        return None
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], X_real_time.shape[2])
    
    predictions = model.predict(X_real_time)
    predictions = scaler.inverse_transform(predictions)
    return predictions

# 4. 에러 처리 및 로그 기록
def get_real_time_data_with_retry(ticker, retries=3, delay=5):
    """서버 오류가 발생할 경우 재시도를 통해 실시간 데이터를 가져옵니다."""
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/inquire-price"
    headers = {
        "Content-Type": "application/json; charset=utf-8",
        "Authorization": f"Bearer {get_valid_token()}",
        "appkey": settings['appkey'],
        "appsecret": settings['appsecret'],
        "tr_id": "FHKST01010100"
    }
    params = {
        "FID_COND_MRKT_DIV_CODE": "J",
        "FID_INPUT_ISCD": ticker
    }
    for attempt in range(retries):
        try:
            response = requests.get(url, headers=headers, params=params)
            response.raise_for_status()
            data = response.json()
            if data['rt_cd'] == '0':
                return pd.DataFrame([data['output']])
            else:
                logging.error(f"{ticker}에 대한 데이터 가져오기 오류: {data['msg1']}")
                return None
        except requests.exceptions.RequestException as e:
            logging.error(f"{attempt + 1}번째 시도 실패: {e}")
            if attempt < retries - 1:
                time.sleep(delay * (2 ** attempt))  # Exponential backoff
            else:
                logging.error(f"{ticker}에 대한 데이터 가져오기 실패: {retries}회 시도 후 실패.")
                return None

def run_websocket():
    """실시간 데이터를 수신하기 위해 WebSocket 클라이언트를 실행합니다."""
    websocket_url = "ws://ops.koreainvestment.com:21000"
    ws = websocket.WebSocketApp(websocket_url,
                                on_open=on_open,
                                on_message=on_message,
                                on_error=on_error,
                                on_close=on_close)
    ws.run_forever()

def is_trading_hours():
    """거래 시간(예: 09:00 ~ 15:30) 여부를 확인합니다."""
    now = datetime.now()
    return now.weekday() < 5 and now.time() >= datetime.strptime("09:00", "%H:%M").time() and now.time() <= datetime.strptime("15:30", "%H:%M").time()

def is_update_time():
    """모델 업데이트가 필요한 시간(오후 6시 이후) 여부를 확인합니다."""
    now = datetime.now()
    return now.time() >= datetime.strptime("18:00", "%H:%M").time()

def batch_collect_data(ticker, start_date, end_date):
    """Batch data collection method."""
    # 기존의 batch_collect_data 코드 유지
    pass

def normalize_data(data):
    """Normalize data."""
    scaler = StandardScaler()
    data['close'] = scaler.fit_transform(data['close'].values.reshape(-1, 1))
    return data, scaler

def preprocess_real_time_data_all(real_time_data):
    """실시간 데이터를 전체 데이터에 맞게 전처리합니다."""
    # 예시: 필요한 모든 특성을 계산하고 real_time_data에 추가합니다.
    # 여기서는 기본적으로 'close', 'volume' 등의 특성만 가정합니다.
    
    # 예: 다른 특성들도 여기서 추가적으로 계산하고 real_time_data에 포함시켜야 합니다.
    
    # 예: 새로운 특성을 추가하는 로직
    # real_time_data['new_feature'] = calculate_new_feature(real_time_data)
    
    # 실제 코드에서 필요한 모든 전처리 과정을 추가하세요.
    
    return real_time_data

def update_lstm_model(y_real_time, model, time_step):
    """Update LSTM model."""
    # Placeholder for model update logic
    return model

def risk_management(predictions, threshold=0.05):
    """Risk management based on predictions."""
    # 기존의 risk_management 코드 유지
    pass

def send_alert(message):
    """Send an alert with a message."""
    # 기존의 send_alert 코드 유지
    pass
# live_data_processing.py

# live_data_processing.py

# stock_data_file_path 변수를 사용하지 않도록 잘못된 코드 제거
# 이 부분은 main_total.py 파일에서 호출하는 방식으로만 사용됩니다.

# 함수 정의만 남깁니다.
def get_stock_codes_from_csv(stock_data_file_path):
    """CSV 파일에서 주식 코드를 읽어옵니다."""
    try:
        if not os.path.exists(stock_data_file_path):
            raise FileNotFoundError(f"CSV 파일을 찾을 수 없습니다: {stock_data_file_path}")
        
        df = pd.read_csv(stock_data_file_path, dtype={'code': str}, usecols=['code'])
        
        if df.empty:
            raise ValueError("CSV 파일이 비어 있습니다.")
        
        return df['code'].tolist()
        
    except Exception as e:
        logging.error(f"CSV 파일에서 주식 코드를 읽는 중 오류가 발생했습니다: {e}")
        return []




# stock_data_file_path가 필요할 때 이 함수를 사용하도록 해야 합니다.
# 이 함수가 올바르게 경로를 받는지 확인하세요.

stock_codes = get_stock_codes_from_csv(stock_data_file_path)  # 예시로 파일에서 불러오는 코드


def fetch_real_time_predictions(stock_code, stock_model, stock_scaler, time_step=60):
    """
    Fetches real-time predictions for a given stock code.

    Args:
    - stock_code (str): Stock code for which predictions are to be made.
    - stock_model: Trained LSTM model.
    - stock_scaler: Scaler used for normalizing the data.
    - time_step (int): The number of previous time steps to consider for prediction.

    Returns:
    - prediction: The predicted value or None if there is an error.
    """
    try:
        # Fetch real-time data
        data = get_real_time_data(stock_code)

        if len(data) < time_step:
            logging.warning(f"{stock_code}에 대한 데이터가 충분하지 않습니다. 필요한 데이터: {time_step}, 받은 데이터: {len(data)}.")
            return None

        # Preprocess the data
        data_scaled = stock_scaler.transform(data[-time_step:])
        data_scaled = data_scaled.reshape((1, time_step, data_scaled.shape[1]))

        # Predict
        prediction = stock_model.predict(data_scaled)
        return prediction
    except Exception as e:
        logging.error(f"Failed to fetch real-time predictions for {stock_code}: {e}")
        return None

def run_trading(stock_codes, stock_model, stock_scaler):
    """
    Runs trading based on real-time predictions.

    Args:
    - stock_codes (list): List of stock codes to trade.
    - stock_model: Trained LSTM model.
    - stock_scaler: Scaler used for normalizing the data.
    """
    try:
        for stock_code in stock_codes:
            prediction = fetch_real_time_predictions(stock_code, stock_model, stock_scaler)

            if prediction is not None:
                # Execute trade based on the prediction
                execute_trade(stock_code, prediction)
            else:
                logging.error(f"Skipping trade execution for {stock_code} due to insufficient data or prediction error.")
    except Exception as e:
        logging.error(f"An error occurred during trade execution: {e}")

# Call run_trading with the list of stock codes
run_trading(stock_codes, stock_model, stock_scaler)

if __name__ == '__main__':
    STOCK_TICKER = "005930"  # 예시: 삼성전자 티커

    stock_model, stock_scaler = None, None
    all_model, all_scaler = None, None

    try:
        # 주식 데이터만 사용하는 모델과 스케일러를 로드
        stock_model, stock_scaler = load_lstm_model_and_scaler(stock_model_path, stock_scaler_path)

        # 전체 데이터를 사용하는 모델과 스케일러를 백그라운드에서 로드 (거래 종료 후 사용)
        all_model, all_scaler = load_lstm_model_and_scaler(all_model_path, all_scaler_path)

        while True:
            if is_trading_hours():
                # 거래 시간 중에는 주식 데이터만 사용하는 실시간 예측을 수행
                if stock_model is None or stock_scaler is None:
                    logging.warning("주식 데이터만 사용하는 모델이 준비되지 않았습니다. 비거래 시간에 모델을 업데이트해야 합니다.")
                else:
                    real_time_predictions = real_time_prediction(stock_model, stock_scaler, STOCK_TICKER)
                    if real_time_predictions is not None:
                        logging.info(real_time_predictions)
                    else:
                        logging.error("Failed to fetch real-time predictions.")
            elif is_update_time():
                # 비거래 시간에 전체 데이터 모델 업데이트 수행 (오후 6시 이후)
                stock_data = batch_collect_data(STOCK_TICKER, start_date=datetime.now() - timedelta(days=7), end_date=datetime.now())
                if stock_data is not None:
                    financial_data = pd.DataFrame()  # Placeholder
                    news_data = pd.DataFrame()  # Placeholder
                    economic_data = pd.DataFrame()  # Placeholder

                    # 데이터 전처리 및 모델 업데이트
                    (X_train, X_test, y_train, y_test), all_scaler, combined_data = preprocess_data(stock_data, financial_data, news_data, economic_data)
                    all_model = train_lstm_model(X_train, y_train)

                    logging.info("전체 데이터를 사용하는 모델이 성공적으로 업데이트되었습니다.")
                else:
                    logging.error("모델 업데이트에 필요한 데이터를 가져오는 데 실패했습니다.")
            else:
                logging.info("거래 시간 외이며 모델 업데이트 시간도 아닙니다. 잠시 대기합니다.")

            # 주기적으로 확인 (예: 1분 간격)
            time.sleep(60)

    except Exception as e:
        logging.error(f"An error occurred: {e}")
