import sys
import os
import pandas as pd
import json
import numpy as np
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import LSTM, Dense, Dropout
from tensorflow.keras.models import load_model as keras_load_model
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping
import logging
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import train_test_split
import joblib

# 경로 설정
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'scripts', 'model')))

# 로그 설정
logging.basicConfig(level=logging.INFO)

# 플래그 설정
SAVE_MODEL_FLAG = True
LOAD_MODEL_FLAG = True

# 경로 설정
MODEL_DIR = "C:/Users/user/projects/st/data/models"
MODEL_PATH_TEMPLATE = os.path.join(MODEL_DIR, "lstm_model_{}.h5")
MODEL_JSON_PATH_TEMPLATE = os.path.join(MODEL_DIR, "lstm_model_{}.json")
SCALER_PATH_TEMPLATE = os.path.join(MODEL_DIR, "scaler_{}.pkl")
PREPROCESSED_DATA_PATHS = {
    "all": "C:/Users/user/projects/st/data/preprocessed_data.csv",
    "stock": "C:/Users/user/projects/st/data/preprocessed_data_stock.csv"
}

def load_preprocessed_data(file_path):
    """Load preprocessed dataset from a file."""
    df = pd.read_csv(file_path)
    X = df.drop(['target', 'date'], axis=1)
    y = df['target']
    return X, y

def build_lstm_model(input_shape, lstm_units=50, dropout_rate=0.2):
    """Build LSTM model."""
    model = Sequential()
    model.add(LSTM(units=lstm_units, return_sequences=True, input_shape=input_shape))
    model.add(LSTM(units=lstm_units, return_sequences=False))
    model.add(Dropout(dropout_rate))
    model.add(Dense(1, activation='linear'))
    model.compile(optimizer='adam', loss='mean_squared_error')
    return model

def train_lstm_model(X_train, y_train, lstm_units=50, dropout_rate=0.2, epochs=20):
    """Train LSTM model."""
    model = build_lstm_model((X_train.shape[1], 1), lstm_units, dropout_rate)

    # 체크포인트 및 조기 종료 콜백 추가
    checkpoint = ModelCheckpoint(filepath=os.path.join(MODEL_DIR, "checkpoint_{epoch:02d}.h5"),
                                 save_freq='epoch',
                                 verbose=1)
    early_stop = EarlyStopping(monitor='loss', patience=5, verbose=1)

    # 모델 학습
    model.fit(X_train, y_train, epochs=epochs, batch_size=32, callbacks=[checkpoint, early_stop], verbose=1)
    
    return model

def save_model(model, model_path, model_json_path, scaler, scaler_path):
    """Save the trained model and its structure."""
    if SAVE_MODEL_FLAG:
        model.save(model_path)
        logging.info(f"Model saved to {model_path}")

        # 모델 구조를 JSON 파일로 저장
        model_json = model.to_json()
        with open(model_json_path, "w") as json_file:
            json_file.write(model_json)
        logging.info(f"Model JSON saved to {model_json_path}")
        
        # 스케일러 저장
        joblib.dump(scaler, scaler_path)
        logging.info(f"Scaler saved to {scaler_path}")
    else:
        logging.info("Model saving skipped due to SAVE_MODEL_FLAG being set to False.")


def load_model(model_path, model_json_path, scaler_path):
    """Load the saved model and its structure."""
    if not os.path.exists(scaler_path):
        logging.error(f"Scaler file not found: {scaler_path}")
        return None, None

    
    if LOAD_MODEL_FLAG:
        model = keras_load_model(model_path)
        logging.info(f"Model loaded from {model_path}")
        
        # JSON 파일 불러오기
        try:
            with open(model_json_path, 'r') as json_file:
                model_json_content = json.load(json_file)
            logging.info("Model JSON loaded successfully.")
        except Exception as e:
            logging.error(f"An error occurred while loading the JSON file: {e}")
        
        # 스케일러 불러오기
        scaler = joblib.load(scaler_path)
        logging.info(f"Scaler loaded from {scaler_path}")
        
        return model, scaler
    else:
        logging.info("Model loading skipped due to LOAD_MODEL_FLAG being set to False.")
        return None, None
    

def update_lstm_model(model, X_train, y_train, epochs=5, batch_size=32):
    """Updates the LSTM model with additional training."""
    model.fit(X_train, y_train, epochs=epochs, batch_size=batch_size, verbose=1)
    return model

def process_and_train(data_type):
    """Process data and train model for a specific data type."""
    file_path = PREPROCESSED_DATA_PATHS[data_type]
    model_path = MODEL_PATH_TEMPLATE.format(data_type)
    model_json_path = MODEL_JSON_PATH_TEMPLATE.format(data_type)
    scaler_path = SCALER_PATH_TEMPLATE.format(data_type)
    
    # 전처리된 데이터 로드
    X, y = load_preprocessed_data(file_path)
    
    # 데이터 분할 및 스케일링
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
    scaler = MinMaxScaler(feature_range=(0, 1))
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)
    
    # 모델 학습 및 저장
    model = train_lstm_model(X_train, y_train, epochs=20)
    save_model(model, model_path, model_json_path, scaler, scaler_path)
    
    # 모델 로드 및 JSON 로드
    loaded_model, loaded_scaler = load_model(model_path, model_json_path, scaler_path)
    if loaded_model:
        logging.info(f"Model for {data_type} loaded and ready for further use.")
    else:
        logging.error(f"Failed to load the model for {data_type}.")


if __name__ == "__main__":
    for data_type in PREPROCESSED_DATA_PATHS:
        process_and_train(data_type)
