import numpy as np
import pandas as pd
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense
from tensorflow.keras.optimizers import Adam

class TradingEnv:
    def __init__(self, data, initial_balance=10000):
        self.data = data
        self.initial_balance = initial_balance
        self.reset()

    def reset(self):
        self.balance = self.initial_balance
        self.position = 0
        self.current_step = 0
        self.done = False
        return self._get_state()

    def _get_state(self):
        state = self.data.iloc[self.current_step].values
        return np.append(state, [self.balance, self.position])

    def step(self, action):
        if self.done:
            return self._get_state(), self.reward, self.done, {}

        self.current_step += 1
        reward = 0

        if action == 1:  # Buy
            if self.balance > self.data.iloc[self.current_step]['Close']:
                self.position += 1
                self.balance -= self.data.iloc[self.current_step]['Close']
        elif action == 2:  # Sell
            if self.position > 0:
                self.position -= 1
                self.balance += self.data.iloc[self.current_step]['Close']

        self.reward = self.balance + self.position * self.data.iloc[self.current_step]['Close'] - self.initial_balance
        self.done = self.current_step >= len(self.data) - 1
        return self._get_state(), self.reward, self.done, {}

def build_dqn_model(input_shape, action_space):
    model = Sequential()
    model.add(Dense(24, input_shape=input_shape, activation='relu'))
    model.add(Dense(24, activation='relu'))
    model.add(Dense(action_space, activation='linear'))
    model.compile(loss='mse', optimizer=Adam(lr=0.001))
    return model

def train_reinforcement_learning_model(stock_data, episodes=100):
    env = TradingEnv(stock_data)
    action_space = 3  # Buy, Hold, Sell
    input_shape = (env.reset().shape[0],)

    model = build_dqn_model(input_shape, action_space)

    for episode in range(episodes):
        state = env.reset()
        total_reward = 0
        done = False
        while not done:
            action = np.argmax(model.predict(state.reshape(1, -1)))
            next_state, reward, done, _ = env.step(action)
            total_reward += reward

            target = reward + 0.95 * np.amax(model.predict(next_state.reshape(1, -1)))
            target_f = model.predict(state.reshape(1, -1))
            target_f[0][action] = target

            model.fit(state.reshape(1, -1), target_f, epochs=1, verbose=0)
            state = next_state

        print(f"Episode: {episode+1}/{episodes}, Total Reward: {total_reward}")

    return model
